%******************************************************************************;
%* SAS macro Crossref vs. 0.8.5 , 03 Feb 2017, by Jim Groeneveld, Netherlands *;
%******************************************************************************;
%* STILL UNDER CONSTRUCTION, NOT FOR VALIDATION, USE WITH CAUTION AT OWN RISK *;
%*                                                                            *;
   %MACRO   Crossref   /* cross reference of multiple datasets                */
/* -------------------------------------------------------------------------- */
/* ---------- A l w a y s   s p e c i f i e d   a r g u m e n t s ----------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
(  DataList= _LAST_    /* input dataset or list of library.dataset names or R */
                       /*    library._ALL_ or library.* for all datasets      */
,  Crossref=           /* output dataset, the resulting cross reference     R */
,  OverWrit= No        /* Yes: overwrite existing CrossRef output dataset   R */
/*                                                                            */
/* ------ F r e q u e n t l y   s p e c i f i e d   a r g u m e n t s ------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
/* Attributes: if empty then no inclusion of concerning attributes, else text */
/* Variable attributes (descriptive values limited to 32 characters)          */
,  Variable='>VARIABLE'/* not empty: include Name and Label  attribute info O */
                       /* if  empty: no variable attributes [and aggregated   */
                       /* data] (Label, Type_Len, Informat, Format, etc.) and */
                       /* variable descriptive statistics (Mean, etc.) either */
,  VarLabel=/*>Label*/ /* not empty: include Variable Label  attribute info O */
                       /* and Variable attribute info in different records:   */
                       /* >Variable   VarName    VarName    VarName    ...... */
                       /* >VarLabel   VarLabel1  VarLabel2  VarLabel3  ...... */
                       /* if empty: Variable Name and Label in same record:   */
                       /* VarName     VarLabel1  VarLabel2  VarLabel3  ...... */
,  Type_Len='>Type/Len'/* not empty: include Type and Length attribute info O */
,  Informat='>Informat'/* not empty: include Informat atribute information  O */
,  Format  ='>Format'  /* not empty: include Format   atribute information  O */
,  VarNum  =           /* not empty: include VarNum   atribute information  O */
/*                                                                            */
/* Numeric and character variable descriptive statistics (max. 32 characters) */
,  NofValid=           /* not empty: include number of Non-missing values   O */
,  Missings=           /* not empty: include number of any Missing values   O */
,  Uniques =           /* not empty: include number of unique valid values  O */
,  Minimum =           /* not empty: include Minimum non-missing value      O */
,  Fminimum=           /* not empty: include formatted Minimum valid value  O */
,  Maximum =           /* not empty: include Maximum non-missing value      O */
,  Fmaximum=           /* not empty: include formatted Maximum valid value  O */
/*                                                                            */
/* Numeric variable descriptive statistics only from non-missing values (<=32)*/
,  Mean    =           /* not empty: include Mean value (unformatted)       O */
,  StDev   =           /* not empty: include Standard Dev. (unformatted)    O */
,  Median  =           /* not empty: include Median value (unformatted)     O */
,  Fmedian =           /* not empty: include formatted Median value         O */
,  Mode    =           /* not empty: include Mode value (unformatted)       O */
,  Fmode   =           /* not empty: include formatted Mode value           O */
/*                                                                            */
/* ---- O c c a s i o n a l l y   s p e c i f i e d   a r g u m e n t s ----- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
/* Attributes: if empty then no inclusion of concerning attributes, else text */
/* Dataset attributes (descriptive values limited to 32 characters)           */
,  LibName ='*Library' /* not empty: include Library Name    attribute info O */
,  MemName ='*Dataset' /* never unspecified: Dataset Name    attribute info R */
,  MemLabel='*DsLabel' /* not empty: include Dataset Label   attribute info O */
,  Original='*Created' /* not empty: include Create datetime attribute info O */
,  Modified='*Changed' /* not empty: include Change datetime attribute info O */
,  Engine  =           /* not empty: include Engine          attribute info O */
,  MemType =           /* not empty: include MemType         attribute info O */
,  TypeMem =           /* not empty: include TypeMem         attribute info O */
,  Compress=           /* not empty: include Compress        attribute info O */
,  NoDupKey=           /* not empty: include NoDupKey        attribute info O */
,  NoDupRec=           /* not empty: include NoDupRec        attribute info O */
,  Sorted  =           /* not empty: include Sorted          attribute info O */
,  SortedBy=           /* not empty: include SortedBy        attribute info O */
,  Charset =           /* not empty: include Charset         attribute info O */
,  Collate =           /* not empty: include Collate         attribute info O */
,  Physical='*Records' /* not empty:include Physical records attribute info O */
,  Logical ='*NofRows' /* not empty: include Logical records attribute info O */
,  NofVars ='*Columns' /* not empty: include N of variables  attribute info O */
,  ColWidth=           /* not empty: include CrossRef column lengthc   info O */
/*                                                                            */
/* ---------- R a r e l y   s p e c i f i e d   a r g u m e n t s ----------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  SQLstats=           /* not empty: include list of aggregate SQL stats    O */
,  UNIstats=           /* not empty: include list of PROC UNIVARIATE stats  O */
,  VarList =           /* variable list specification to include only       O */
,  ComMinNr= /*1=all*/ /* Minimum Number of occurrences of Common variables O */
                       /*   (Default 1) in involved datasets for inclusion    */
                       /*   in the cross reference dataset (see note *).      */
,  ComMaxNr= /*0=all*/ /* Maximum Number of occurrences of Common variables O */
                       /*   (default 0) in involved datasets for inclusion    */
                       /*   in the cross reference dataset (see note *).      */
,  CalledBy=CrossRef   /* Macro name of macro or program for err0r report   O */
,  Version = 0.8.5     /* forced version specification for version control  O */
/*                                                                            */
/* ----------- N e v e r   s p e c i f i e d   a r g u m e n t s ------------ */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  Debug   = 0         /* Debug level; for testing, debugging purposes only O */
,  ForMyUse=           /* argument for the program's use only, don't use it O */
)/ DES     =              'cross reference of multiple datasets' /*Store*/     ;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* File       : Crossref.SAS / Crossref.084                                   *;
%*                                                                            *;
%* Purpose    : Cross reference of data structure of multiple datasets        *;
%*              by common variable names of existing variables in datasets    *;
%*              involving various (optional) dataset and variable attributes  *;
%*              and numeric univariate statistics.                            *;
%*              Dataset attributes include Dataset Library, Name and Label,   *;
%*              Creation (CrDate) and Modification (MoDate) dates,            *;
%*              Engine, TypeMem, MemType, Compress, NoDupKey, NoDupRec,       *;
%*              Sorted, SortedBy, Charset, Collate,                           *;
%*              number of Physical and Logical records, number of Variables.  *;
%*              Variable attributes include Name and Label, VarNum, Type and  *;
%*              Length, Informat and Format, number of Non-missing, Missing   *;
%*              and Unique (non-missing) values, (non-missing) Minimum and    *;
%*              Maximum statistics (both unformatted and formatted).          *:
%*              Numeric variable univariate statistics include Mean and       *;
%*              St.Dev., Median and Mode (both unformatted and formatted)     *;
%*              and _all_ (other) SQL aggregate function statistics and       *;
%*              PROC UNIVARIATE summary statistics (see SAS documentation.).  *;
%*              The result is a cross reference dataset with an Attribute     *;
%*              column and one column for each involved dataset (or view)     *;
%*              that can be presented as desired (e.g. using PROC REPORT).    *;
%*                                                                            *;
%* Input      : (related) dataset name list and/or _ALL_ (or *) specification *;
%*                                                                            *;
%* Output     : cross reference inventory of all involved datasets            *;
%*                                                                            *;
%* Programmer : Jim Groeneveld                                                *;
%* Date       : 03 Feb 2017                                                   *;
%* Version    : 0.8.5                                                         *;
%*                                                                            *;
%* SAS version: 9.1.3 and 9.2 under Windows XP                                *;
%*                                                                            *;
%* Validation : -                                                             *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Description                                                                *;
%* -----------                                                                *;
%* paper Comparing dataset metadata, PhUSE 2011, Brighton, October 9-12, 2011 *;
%* http://www.phuse.eu/download.aspx?type=cms&docID=3602                      *;
%* All 'future' features, described in there have actually been implemented.  *;
%*                                                                            *;
%* Notes                                                                      *;
%* -----                                                                      *;
%* *) ComMinNr and ComMaxNr arguments - firstly ComMinNr (ComMaxNr: see e.):  *;
%* a. Default minimum number = 1: all variables involved. Example: 2 or 3     *;
%*    causes to include only variables that occur at least in 2 or 3 datasets.*;
%* b. Zero or a negative value indicates the number of datasets minus the     *;
%*    absolute (positive) value of the number. E.g. -2 with 8 datasets means  *;
%*    a minimum number of 6 occurrences of a variable in the 8 datasets.      *;
%* c. A positive number with a percent sign (%) indicates a minimum number    *;
%*    relative to the number of datasets. E.g. 50% with 8 datasets resolves   *;
%*    to a minimum number of 4 occurrences of a variable in the 8 datasets.   *;
%* d. 0 or 100% thus means to include variables if they occur in all datasets.*;
%*    1 (default) or 0% indicates to always include all variables. Empty = 1. *;
%* e. The same applies to ComMaxNr for the maximum number of occurrances.     *;
%* f. To output only identical structures with regard to variable names use   *;
%*    ComMinNr=0 or 100% (and ComMaxNr as well, default).                     *;
%*    To output only non-identical structures with regard to variable names   *;
%*    use ComMaxNr=-1 (and ComMinNr=1, default).                              *;
%*                                                                            *;
%* Download location                                                          *;
%* -----------------                                                          *;
%* http://home.hccnet.nl/jim.groeneveld/software/SASmacro/CrossRef.zip        *;
%* http://home.hccnet.nl/jim.groeneveld/crossref                              *;
%*                                                                            *;
%* Limitations                                                                *;
%* -----------                                                                *;
%* none (those of version 0.6.3. have been removed, see History)              *;
%*                                                                            *;
%* Remarks                                                                    *;
%* -------                                                                    *;
%* This macro has not been tested under Unix/Linux.                           *;
%* The %UPCASE and UPCASE functions should be removed/deactivated under Unix. *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Example macro call                                                         *;
%* ------------------                                                         *;
/* %Crossref (DataList=DataLib.*, Crossref=Library.Crossref, Overwrit=Y,      */
/*                     Engine='*Engine', MemType='*MemType',                  */
/*                     TypeMem='*TypeMem', Compress='*Compres',               */
/*                     NoDupKey='*NDupKey', NoDupRec='*NDupRec',              */
/*                     Sorted='*Sorted', SortedBy='*SortdBy',                 */
/*                     Charset='*Charset', Collate='*Collate',                */
/*                     VarNum='>OrderNr', Uniques='>Distinct',                */
/*                     NofValid=">ValidNum", Missings=">Missings",            */
/*                     Minimum=">Lowest", FMinimum=">fLowest",                */
/*                     Maximum=">Highest", FMaximum=">fHighest",              */
/*                     Mean=">Average", StDev=">St. Dev.",                    */
/*                     Median=">Median50", Fmedian=">Fmedian50",              */
/*                     Mode=">HighFreq", Fmode=">FhighFreq",                  */
/*                     SQLstats=Range Sum, UNIstats=Cv Var,                   */
/*                     ColWidth='#FieldLen', VarList=Var1 Var2,               */
/*                     ComMinNr=50%, CalledBy=SAS_code, Debug=0);             */
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Used auxiliary macros:                                                     *;
%* ----------------------                                                     *;
%* = internal: -                                                              *;
%* = attached: DescStat                                                       *;
%* = external: %UstrMcmp, %ERR0R, %WARN1NG, %_Delete_                         *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* History of Crossref:                                                       *;
%* ====================                                                       *;
%* version  date         description of changes, additions, improvements      *;
%* -------  -----------  ---------------------------------------------------- *;
%* 0.0.0a   13 Oct 2005  first Q(uick)&D(irty) working release named VarLists *;
%* 0.0.0b    5 Aug 2011  capitalised variable Names to match equal variables  *;
%*                          with names in different case in the datasets      *;
%* 97a1/0.0 10 Jul 2009  working version with variable labels,          `     *;
%*                       macro processed just one dataset, results had to be  *;
%*                          combined in dedicated code outside macro          *;
%*                       -->changed name of this macro to MetaData            *;
%* 97e1/0.1 14 Jul 2009  added number of records as additional record,        *;
%*                       changed name sorting into case-independent,          *;
%* 97f1     15 Jul 2009  changed some experimental code back to normal        *;
%* 97f2/0.2 15 Jul 2009  added Date & Time of creation and last               *;
%*                          modification, added number of variables           *;
%* 97g1/0.3 16 Jul 2009  added Dataset Label, changed name of created         *;
%* 97r1/0.4 27 Jul 2009  replaced APPEND by SET due to absent vars            *;
%* 97s1  28-30 Jul 2009  transferred labels from original datasets,           *;
%*                          else added informative text labels,               *;
%*                       transferred SAS time from original datasets          *;
%*                          or Windows time with non-SAS files, Excel (test)  *;
%* 9c91     09 Dec 2009  (added two more datasets to compare while testing)   *;
%* a1b1/0.5 12 Jan 2010  (added another dataset to compare while testing)     *;
%* 0.6.0     3 Aug 2011  rewritten all comment and macro layout,          `   *;
%*                       -->changed name of this macro to CrossRef            *;
%* 0.6.1     4 Aug 2011  made suited for a DataList of dataset specifications *;
%*                          (so combining all datasets now part of the macro),*;
%*                       added _LibName attribute,                            *;
%*                       changed _Columns into _NofVars,                      *;
%*                       changed _Records into _NofObs (physical records),    *;
%*                       added optional attribute information via arguments   *;
%*                          TypeLen, Informat and Format,                     *;
%*                       capitalised variable names like in vs. 0.0.0b.       *;
%*                          This actually was a severe ommission in previous  *;
%*                          prototype versions 0.0 to 0.5, but not fatal,     *;
%* 0.6.2     9 Aug 2011  added argument Overwrit to let the user decide,      *;
%*                       changed experimental character into final macro by:  *;
%*                       - adding and improving documentation and description,*;
%*                       - making debugging code dependent from Debug arg.,   *;
%*                       - replacing work dataset names by macro variables    *;
%*                         with defined values starting with __,              *;
%*                       - storing and restoring all altered OPTION settings, *;
%*                       - deletion of work datasets at the end of the macro  *;
%* 0.6.3    12 Aug 2011  NofLogObs and NofDelObs directly obtained from PROC  *;
%*                          CONTENTS output. NofPhysObs=NofLogObs+NofDelObs,  *;
%*                       NofPhysObs obtained from calculation above instead   *;
%*                          of from dataset header,                           *;
%*                       changed name of argument TypeLen into Type_Len,      *;
%*                       made all (dataset and variable) attributes optional  *;
%*                          by adding macro parameters for them with text     *;
%*          15 Aug 2011  corrected buggy LibName for WORK datasets,           *;
%*                       changed sorting datasets and variable attributes at  *;
%*                          the end of the macro,                             *;
%*                       added arguments NofVars, Physical and Logical,       *;
%*                          making them optional                              *;
%*           6 Mar 2012  fixed buggy determination of Formats (0.6.1 & 0.6.2) *;
%* 0.7.0     8 Feb 2012  largely rewritten algorithm, without internally      *;
%*                          using SAS name literal variable names, removing   *;
%*                          limitation to dataset variables of 30 characters, *;
%*                       removed limitation not to allow variable names equal *;
%*                          to attribute names. It is, however, recommended   *;
%*                          to avoid such visually confusing variable names.  *;
%*                       This version is functionally equivalent to version   *;
%*                       -->0.6.3, but the column names are numbered datasets *;
%*                          (was: dataset names, problem: same in other lib.),*;
%*                          the column label now is 'Libname.Dataset' (in     *;
%*                          specified case (but in attributes as capitals))   *;
%*                          and the dataset`s number of variables is listed   *;
%*                          below the number of physical and logical records  *:
%*                          (was: just above). Minor improvements in comments *;
%*                       Fixed bug with _LAST_ as dataset/list specification  *;
%* 0.7.1    13 Feb 2012  Added support for dataset specification _ALL_ or *   *;
%*                          implicitly including all datasets in library      *;
%*                          (dataset column labels now are in capitals too)   *;
%* 0.7.2    14 Feb 2012  Added additional dataset attribute information:      *;
%*                          arguments VarNum, Engine, MemType, TypeMem,       *;
%*                          NoDupKey, NoDupRec, Sorted, SortedBy, Charset     *;
%*                          and Collate                                       *;
%* 0.7.3     6 Mar 2012  Added arguments displaying descriptive statistics:   *;
%*                          num+char: NoMisses, Missings, Minimum, Fminimum,  *;
%*                                    Maximum, Fmaximum                       *;
%*                          num only: Mean, StDev (all PROC SQL aggr. stats)  *;
%*                       Made column widths dependent on data (desc. values), *;
%*                       Changed argument Name_Lbl into Variable              *;
%* 0.7.4     6 Mar 2012  Added argument Median (using PROC UNIVARIATE) as     *;
%*                          the median does not exist as a PROC SQL aggr.     *;
%*                          stat. Likewise many more statistics may be used.  *;
%*           7 Mar 2012  Added argument Fmedian (formatted median value),     *;
%*           8 Mar 2012  Added arguments Uniques, Mode and (formatted) Fmode, *;
%*                          ColWidth (width of column in CrossRef dataset),   *;
%*                       Internal optimisations (process lists of attributes) *;
%* 0.7.5     9 Mar 2012  WORK library specification for one-level dataset     *;
%*                          specifications obtained from System Option WORK   *;
%*                          using GETOPTION instead of just assuming WORK,    *;
%*                       Added arguments SQLstats and UNIstats to specify any *;
%*                          list of SQL aggregate or PROC UNIVARIATE summary  *;
%*                          statistics. The user should exactly specify the   *;
%*                          appropriate statistics by name. These names are   *;
%*                          preceded by an underscore in the CrossRef dataset.*;
%* 0.7.6    12 Mar 2012  Added argument VarList (variable list) to limit the  *;
%*                          comparison to user specified variables.           *;
%* 0.7.7    21 Mar 2012  Fixed bug with non-specified ColWidth argument from  *;
%*                          vs. 0.7.4, 0.7.5 and 0.7.6 .                      *;
%*          23 Mar 2012  fixed bug with NofVars results on CrossRef since     *;
%*                          vs. 0.7.1 in case DataList contains _ALL_ or *    *;
%* 0.7.8    15 May 2012  fixed bug with specified VarList generating VarName  *;
%*                          since vs. 0.7.6                                   *;
%*          17 May 2012  fixed bug with (in)formats without widths & decimals *;
%*                          since vs. 0.6.3 (or rather all previous versions) *;
%* 0.7.9    17 May 2012  Added argument ComMinNr, indicating the Minimum      *;
%*                          Number of Common occurrences of a variable in     *;
%*                          datasets for inclusion in the CrossRef dataset,   *;
%*                          including the interpretation of zero and negative *;
%*                          values or values specified as a percentage.       *;
%* 0.8.0     2 Aug 2012  fixed bug not outputting Fmedian and Fmode without   *;
%*                          Median and Mode                                   *;
%* 0.8.1     8 Oct 2012  fixed bug with existing and long format names        *;
%* 0.8.2    22 Jan 2014  fixed bug with WORK library, removed Spacing0,       *;
%*                       fixed bug with requesting no descriptive statistics, *;
%*                       fixed bug with CAPITALISED variable names from PROC  *;
%*                          CONTENTS in specific SAS versions (9.1.z),        *;
%*                       added CalledBy argument                              *;
%* 0.8.3    27 Jun 2014  fixed severe bug with arguments Missings and Uniques *;
%*                          (from vs. 0.7.4) (not showing while specified),   *;
%*                       changed argument name NoMisses into NofValid         *;
%* 0.8.4    30 Jun 2014  added argument VarLabel to show Variable Labels in   *;
%*                          a separate record just below the Variable Names,  *;
%*                          indicated by the user specified VarLabel attr.    *;
%*                          The corresponding, identical Variable Names then  *;
%*                          are given for every dataset in the above record,  *;
%*                          indicated by the user specified Variable attr.    *;
%*                          If VarLabel not specified then Labels presented   *;
%*                          in a record indicated by the Variable Name (old). *;
%*                          This introduces an alternative structure: Names   *;
%*                          and Labels not next to each other but below each  *;
%*                          other with attributes ">Variable" and ">VarLabel".*;
%*                          If both empty then no Variable Names, Labels and  *;
%*                          any other (specified) attributes at all given.    *;
%* 0.8.5      3 Feb 2017  added argument ComMaxNr, similar to ComMinNr,       *;
%*                        fixed bug with VarList, ComMinNr (and ComMaxNr)     *;
%*                          without specified data statistics (only metadata),*;
%*                          a bug that must have existed since version 0.7.6  *;
%*                                                                            *;
%* Future of Crossref: (ideas, wishes, plans - if feasible)                   *;
%* ===================                                                        *;
%* - argument PathName or Folder or SubDir or SubMap for the physical,        *;
%*   OS-dependent, directory name of the library of a dataset (see            *;
%*   http://www.caloxy.com/papers/78WholePath.pdf )                           *;
%* - (even more single variable statistical analyses (via dedicated code))    *;
%* - dataset options (already experimentally programmed separately)           *;
%* - ColWidth actually is the same for each dataset: report once somewhere    *;
%* - test illegal argument input                                              *;
%* - next to limiting the number of variables presented from VarList          *;
%*   and ComMinNr also report the remaining number of variables               *;
%*   per dataset presented as keyword NofNames or in/after NofVars            *;
%*   (if ComMinNr = 1 or 0% and ComMaxNr = 0 or 100%) ?????????               *;
%* - present for each variable at the right side an additional column         *;
%*   with number of common variables over datasets with column                *;
%*   name=CommonNr and label="Number of common datasets" (for variable)       *;
%*   1x per variable in Name/Label row                                        *;
%* - (optionally) remove dataset from CrossRef if no variables involved       *;
%*   from VarList specification or ComMinNr value NE 1, argument ...          *;
%* - calculate percentage non-missings of total and present optionally        *;
%*   (if NoMisses and Missings calculated) keyword: ValidPct                  *;
%* - optional only differences with dataset # [Compare=Dataset#]              *;
%* - or among all datasets (Compare=Datasets)                                 *;
%* - optional Diff_Var indicating highlighting in PROC REPORT                 *;
%*   Diff_Var=0/1 or number of differences per attribute                      *;
%* - check cleaning up, especially options.                                   *;
%* - two optional ordering variables _AttSet_ and _AttNum_. For use with      *;
%*   e.g. PROC REPORT to force order with GROUP variables (NOPRINT) and       *;
%*   for layout purposes, e.g. blank lines in between after last _AttSet_     *;
%*   (BREAK or COMPUTE). Also adapt macro _Report_ to these features.         *;
%*   _AttSet_ = 0 for all Dataset attributes                                  *;
%*   _AttSet_ = 1..NofVars for all Variable attributes, incrementing 1 per    *;
%*              variable                                                      *;
%*   _AttNum_ = unique sequence number of each attribute within a set /       *;
%*              / variable, not necessarily consecutive (each attribute has   *;
%*              fixed number). SQLstats and UNIstats have single large number,*;
%*              e.g. 98 and 99 (optionally to specify when requesting those   *;
%*              stats).                                                       *;
%* - optionally order by VarNum, introduce argument VarOrder                  *;
%*                                                                            *;
%******************************************************************************;

%* OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
%* OPTIONS SPOOL;

  %LOCAL    /* Define internal macro variables explicitely local */
    MacName  /* This macros external name: Crossref */
    MacVs    /* Version number major.minor.patchSub of macro: 0.8.5 */
    MacDate  /* Date of this version of macro: 03 Feb 2017 */
    ErrCount /* number of erroneously specified arguments */
    ArgList  /* List of user specifiable macro arguments */
    Argument /* Single user specifiable macro argument */
    LineSize /* stored LINESIZE option */
    PageSize /* stored PAGESIZE option */
    NoCenter /* stored CENTER option */
    ValidVar /* stored VALIDVARNAME option */
    NoMPrint /* stored MPRINT option */
    NoMErr0r /* stored MERR0R option */
    NoSErr0r /* stored SERR0R option */
    NoMLogic /* stored MLOGIC option */
    Symbolgn /* stored SYMBOLGEN option */
    MacroGen /* stored MACROGEN option */
    NoSpool  /* stored SPOOL option */
    Contents /* output dataset of PROC CONTENTS */
    AttrData /* name of dataset with variable attributes: Type, Length, Informat, Format */
    Appended /* (horizontally structured) dataset to append metadata to */
    I        /* incremental counter for various purposes */
    DataFile /* Dataset element of DataList argument: libname.dataset */
    DataLib  /* Library name of DataFile (for splitting Lib and Set) */
    DataSet  /* Dataset name of DataFile (for splitting Lib and Set) */
/*  Aborting    indication to abort macro */
    DatasetN /* number of requested dataset attributes (in contrast to variable attributes) */
    NofDsets /* number of specified datasets */
    DsetInfo /* name of temporary dataset with dataset info from CONTENTS */
    MetaDset /* name of temporary dataset with dataset metadata */
    MetaVars /* name of temporary dataset with variable metadata */
    DsAttrs  /* List of dataset attributes (in exact case) */
    DsAttr   /* Single variable attribute (in exact case) */
    VarAttrs /* List of variable attributes and statistics (in exact case) */
    VarAttr  /* Single variable attribute or statistics (in exact case) */
    DescStat /* indication of any descriptive statistics requested */
    SQL_stat /* name of temporary dataset with aggregate stats from PROC SQL */
    StatDesc /* name of temporary dataset with statistical descriptors */
    Uni_stat /* name of temporary dataset with UNIVARIATE statistics */
    SQLstat  /* Single aggregate statistic from PROC SQL */
    UNIstat  /* Single summary statistic from PROC UNIVARIATE */
    Var_List /* name of temporary dataset with specified variable names */
    VarName  /* Single variable name from optional variable list */
    NofNames /* Number of variable names left after reducing to conditions */
    ;

  %LET MacName  = Crossref;
  %LET MacVs    = 0.8.5 ;
  %LET MacDate  = 03 Feb 2017;
  %LET LineSize = %SYSFUNC(GETOPTION(LINESIZE));
  %LET PageSize = %SYSFUNC(GETOPTION(PAGESIZE));
  %LET NoCenter = %SYSFUNC(GETOPTION(CENTER));
  %LET ValidVar = %SYSFUNC(GETOPTION(VALIDVARNAME));
  %LET NoMPrint = %SYSFUNC(GETOPTION(MPRINT));
  %LET NoMErr0r = %SYSFUNC(GETOPTION(MERROR));
  %LET NoSErr0r = %SYSFUNC(GETOPTION(SERROR));
  %LET NoMLogic = %SYSFUNC(GETOPTION(MLOGIC));
  %LET Symbolgn = %SYSFUNC(GETOPTION(SYMBOLGEN));
  %LET MacroGen = %SYSFUNC(GETOPTION(MACROGEN));
  %LET NoSpool  = %SYSFUNC(GETOPTION(SPOOL));
  %LET Contents = __ContDt;
  %LET AttrData = __AttrDt;
  %LET Appended = __Append;
  %*LET Aborting = 0;
  %LET DsetInfo = __DsInfo;
  %LET MetaDset = __MetaDs;
  %LET MetaVars = __MetaVs;
  %LET DescStat = %LENGTH (&NofValid &Missings &Uniques
                           &Minimum &Fminimum &Maximum &Fmaximum
                           &Mean &StDev &Median &Fmedian &Mode &Fmode);
  %LET SQL_stat = __SQLSta;
  %LET StatDesc = __StDesc;
  %LET Uni_stat = __UnivSt;
  %LET Var_List = __VarLst;

 %IF ( %UPCASE(&ForMyUse) = COMPILE ) %THEN %GOTO Finish;

%IF (&Debug EQ 1 OR &Debug GT 11) %THEN %DO;
  OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN SPOOL;
%END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Macro header in LOG file;
%*_________________________;

  %PUT;
  %PUT |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|;
  %PUT |   Macro &MacName, vs. &MacVs, by Jim Groeneveld, &MacDate  |;
  %PUT |    Cross reference of data structure of multiple datasets    |;
  %PUT |______________________________________________________________|;
  %PUT;

%* Feedback of specified arguments, programmed as macro loop;
  %PUT &MacName: Review of argument specifications (incl. defaults);
  %LET ArgList = DataList CrossRef Overwrit ComMinNr ComMaxNr LibName MemName MemLabel
                 Original Modified Engine MemType TypeMem Compress NoDupKey
                 NoDupRec Sorted SortedBy Charset Collate Physical Logical
                 NofVars ColWidth Variable VarLabel Type_Len InFormat Format VarNum
                 NofValid Missings Uniques Minimum Fminimum Maximum Fmaximum
                 Mean StDev Median Fmedian Mode Fmode SQLstats UNIstats VarList
                 CalledBy Version Debug;
  %LET I = 1;
  %LET Argument = %SCAN ( &ArgList, &I, %STR( ) );
  %DO %WHILE (%LENGTH(&Argument));
    %PUT &MacName: &Argument = &&&Argument;
    %LET I = %EVAL ( &I + 1 );
    %LET Argument = %SCAN ( &ArgList, &I, %STR( ) );
  %END;

  %IF (&Version NE AND &Version NE &MacVs) %THEN
  %DO;
    %PUT *** &MacName *** %ERR0R-: Specified version &Version does not match &MacVs,;
    %PUT .                       macro &MacName will abort;
    %GOTO Exit;
  %END;

%* Argument checks;

%* DataList argument;
  %IF (NOT %LENGTH(%SUPERQ(DataList))) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: DataList not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

  %IF ( NOT %UstrMcmp(&Overwrit,Yes) AND NOT %UstrMcmp(&Overwrit,No) ) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Overwrit argument not specified or illegal: &Overwrit,;
    %PUT .                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

  %IF ( %SYSFUNC(EXIST(&CrossRef)) AND NOT %UstrMcmp(&OverWrit,Yes) ) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Dataset CrossRef=&CrossRef exists and OverWrit is not Yes,;
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

  %IF (NOT %LENGTH(%SUPERQ(MemName))) %THEN %DO;
    %PUT *** &MacName *** %WARN1NG-: MemName specified empty but is required,;
    %PUT .                     MemName set to default '*Dataset';
    %LET MemName = '*Dataset';
  %END;

%* VarList cannot (easily) be checked for existing variable names;

%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
%* ComMinNr: positive integer=absolute number or percentage (%)        ;
%*       non-positive integer=number of datasets minus absolute number ;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
  %IF (NOT %LENGTH(%SUPERQ(ComMinNr))) %THEN %LET ComMinNr = 1; %* default 1;
  %ELSE %IF (%SYSFUNC(VERIFY(%BQUOTE(&ComMinNr),-0123456789%STR(%%)))) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Illegal value ComMinNr=%BQUOTE(&ComMinNr) (see documentation);
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (%QSUBSTR(&ComMinNr,1,1) EQ %STR(-) AND
             %QSUBSTR(&ComMinNr,%LENGTH(%SUPERQ(ComMinNr)),1) EQ %STR(%%)) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Illegal value ComMinNr=%BQUOTE(&ComMinNr) (negative percentage);
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
%* ComMaxNr: positive integer=absolute number or percentage (%)        ;
%*       non-positive integer=number of datasets minus absolute number ;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
  %IF (NOT %LENGTH(%SUPERQ(ComMaxNr))) %THEN %LET ComMaxNr = 0; %* default 0;
  %ELSE %IF (%SYSFUNC(VERIFY(%BQUOTE(&ComMaxNr),-0123456789%STR(%%)))) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Illegal value ComMaxNr=%BQUOTE(&ComMaxNr) (see documentation);
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (%QSUBSTR(&ComMaxNr,1,1) EQ %STR(-) AND
             %QSUBSTR(&ComMaxNr,%LENGTH(%SUPERQ(ComMaxNr)),1) EQ %STR(%%)) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Illegal value ComMaxNr=%BQUOTE(&ComMaxNr) (negative percentage);
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;

  %IF ( &ErrCount GT 0 ) %THEN %GOTO Aborting;

%* - - - - - - - - - - - ;
%* Core code starts here ;
%* - - - - - - - - - - - ;

**OPTIONS VALIDVARNAME=ANY;

/*PROC DATASETS LIBRARY=WORK; DELETE &Appended; RUN; * delete existing, old result;*/
  DATA &Appended; STOP; RUN; * empty intermediate result dataset as a start ;

  %LET NofDsets = 0;
  %LET I = 1;
  %LET DataFile = %SCAN ( &DataList, &I, %STR( ) );
  %DO %WHILE ( %LENGTH ( %SUPERQ ( DataFile ) ) );

%* Process keyword _LAST_ with DataFile;
    %IF (%QUPCASE(&DataFile) EQ _LAST_) %THEN %LET DataFile = &SYSLAST;

%* Process multiple dataset specifications from keyword _ALL_ or * ;
    %IF (%LENGTH ( %SCAN ( &DataFile, 2, . ))) %THEN %DO;
      %LET DataLib = %SCAN ( &DataFile, 1, . );
      %LET Dataset = %SCAN ( &DataFile, 2, . );
    %END;
    %ELSE %DO;
      %LET DataLib = WORK;
      %LET Dataset = %SCAN ( &DataFile, 1, . );
    %END;

    %IF (%SUPERQ(DataSet) EQ %STR(*)) %THEN %DO;
      %LET DataSet = _ALL_;
      %LET DataFile = &DataLib.._ALL_;
    %END;

%* Check existence of specified dataset;
    %IF (%UPCASE(&DataSet) NE _ALL_) %THEN %DO;
      %IF (NOT %SYSFUNC(EXIST(&DataFile))) %THEN %DO;           
        %PUT *** &MacName *** %ERR0R-: Dataset &DataFile does not exist,;
        %PUT .                     macro &MacName will abort;
        %GOTO Aborting;
      %END;
    %END;

    %LOCAL Dataset;
    %LET Dataset = %SCAN ( &DataFile, -1, .);

    PROC CONTENTS DATA=&DataFile NOPRINT OUT=&Contents; RUN;

%if &Debug ge 1 %then %do; %* = = = = = = = = = ;
  data current.Contents_&I; set &Contents; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = ;

    DATA &AttrData  (KEEP=LibName MemName MemLabel Name Label
                          CrDate MoDate Type Type_Len InFmt Fmt VarNum
                          Nobs Nrecords /*NofVars*/ DatasetN DataFile
                          Engine MemType TypeMem Compress NoDupKey
                          NoDupRec Sorted SortedBy Charset Collate
                          RENAME=(CrDate=Original MoDate=Modified
                          Nobs=Logical Nrecords=Physical
                          InFmt=InFormat Fmt=Format));
* Reorder variables such that VarNum comes after Label;
      RETAIN LibName MemName MemLabel CrDate MoDate Engine MemType TypeMem
             Compress NoDupKey NoDupRec Sorted SortedBy CharSet Collate
            Nrecords NOBS /*NofVars*/ Name Label VarNum; 
      SET &Contents (KEEP=LibName MemName MemLabel Name Label Type Length
                          Informat InformL InformD Format FormatL FormatD
                          CrDate MoDate Nobs DelObs VarNum
                          Engine MemType TypeMem Compress NoDupKey
                          NoDupRec Sorted SortedBy CharSet Collate) NOBS=NVars;
      BY LibName MemName;

****  NofVars = NVars; * (Nvars is a temporary variable);
* - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
* NofVars is total number of records in Contents dataset ;
* If this dataset contains metainfo of more datasets at  ;
* once (from _ALL_) then the number of variables per     ;
* dataset is not correct and will be recalculated below  ;
* using SQL. Keep NofVars in the ordered RETAINed list.  ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - -;

      LENGTH Type_Len InFmt Fmt $40 /* including W, period and D */ DataFile $41;
      Name = UPCASE(Name); * force upper case for matching later;
* missing MemLabel and Label;
      IF (MISSING(MemLabel)) THEN MemLabel = '-no label-';
      IF (MISSING(   Label)) THEN    Label = '-no label-';
* Type and Length;
      IF (Type EQ 1) THEN Type_Len = 'Numeric/8';
      ELSE Type_Len = 'Char/' || LEFT(PUT (Length, BEST.));
* Informat and Format;
      IF (InformL IN (0 .) AND InFormat EQ '') THEN InFmt = '';
      ELSE IF (InformL IN (0 .)) THEN InFmt = TRIM(InFormat) || '.';
      ELSE InFmt = TRIM(InFormat) || TRIM(LEFT(PUT(InFormL,BEST.))) || '.' || TRIM(LEFT(PUT(InFormD,BEST.)));
      IF (FormatL IN (0 .) AND   Format EQ '') THEN Fmt = '';
      ELSE IF (FormatL IN (0 .)) THEN Fmt = TRIM(Format) || '.';
      ELSE Fmt = TRIM(  Format) || TRIM(LEFT(PUT(FormatL,BEST.))) || '.' || TRIM(LEFT(PUT(FormatD,BEST.)));

* Store dataset number with contents info as well;
      RETAIN DatasetN &NofDsets DataFile;
      IF (FIRST.MemName) THEN DO;
        DatasetN + 1;
* Make LibName.MemName combination for labeling dataset columns later;
        DataFile = TRIM(LEFT(LibName)) || '.' || TRIM(LEFT(MemName));
****    NofVars = 0;
      END;
      IF (LAST.MemName) THEN CALL SYMPUT ('NofDsets', PUT(DatasetN, BEST12.));
* Calculate physical number of records from (logical) Nobs and DelObs;
      Nrecords = Nobs + DelObs;
******  NofVars + 1;
    RUN;

    DATA &Appended; SET &Appended &AttrData; RUN; * append contents;

%if &Debug ge 1 %then %do; %* = = = = = = = = = ;
  data current.AttrData_&i; set &AttrData; run; * debug;
  data current.Appended_&i; set &Appended; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = ;

    %_Delete_ (&Contents &AttrData);

    %LET I = %EVAL ( &I + 1 );
    %LET DataFile = %SCAN ( &DataList, &I, %STR( ) );
  %END; %* macro DO WHILE - for each specified dataset or wildcard;

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
* All dataset information is now contained in the dataset &Appended.;
* Extract dataset meta information for each dataset and store it in ;
* a temporary output dataset MetaDset, one record per attribute     ;
* (LibName, MemName, MemLabel, etc.) for all involved datasets.     ;
* Sort the dataset by user specified dataset order and keep one     ;
* record per dataset.                                               ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

* = = = = = = = = = ;
* Dataset metadata  ;
* = = = = = = = = = ;

* Calculate number of variables for each dataset;
  %***IF (%LENGTH(%SUPERQ(NofVars))) %THEN %DO;
    PROC SQL;
      CREATE TABLE &Appended AS                /* original data */
        SELECT * FROM &Appended AS A
          FULL JOIN
            (SELECT N(DatasetN) AS NofVars     /* frequency of  */
              FROM &Appended AS B              /* obs per dataset */
                GROUP BY DatasetN)
                  ON A.DatasetN EQ B.DatasetN; /* merge by datasetN */
    QUIT;
  %***END; %* NofVars has to be present for the below KEEP and DROP lists;

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
* Sort &Appended dataset by MemName and LibName into &MetaDset. ;
* (NODUPKEY option retains the first observation with unique By ;
* variable values, that is why the above step had to be added.) ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

  PROC SORT DATA=&Appended (KEEP=LibName -- Logical NofVars DataFile DatasetN)
            OUT =&DsetInfo NODUPKEY; BY DatasetN; RUN;

%if &Debug ge 1 %then %do; %* = = = = = = = = = ;
  data current.DsetInfo; set &DsetInfo; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = ;

* Extract (optional) dataset metadata from TRANSPOSEd DsetInfo;
  PROC TRANSPOSE DATA=&DsetInfo OUT =&MetaDset (DROP=_LABEL_)
                 NAME=Attribute PREFIX=Dataset_;
* ID statement not necessary itself, but needed to define IDLABEL;
    ID DatasetN; IDLABEL DataFile;
    VAR _ALL_; * Numeric and character variables, incl. DatasetN and DataFile;
* The VAR statement determines the order of Attribute records;
  RUN;

%if &Debug ge 1 %then %do; %* = = = = = = = = = ;
  data current.MetaDset_Transposed; set &MetaDset; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = ;

* Process MetaDset with regard to optional attributes and their names;
  DATA &MetaDset;
    RETAIN Attribute
    %DO I = 1 %TO &NofDsets;
      Dataset_&I
    %END;
    ; * Define order of variables with RETAIN;
    LENGTH Attribute $32; * (must also contain complete variable names);
    SET &MetaDset;
    AttrPart = 'Dataset';
*###Attribute = UPCASE(Attribute); * (always under earlier SAS versions);
    LABEL Attribute = 'columns=datasets, rows=variables and attributes';

* Delete some records that were generated by TRANSPOSing VAR _ALL_;
    IF (UPCASE(Attribute) IN ('DATAFILE' 'DATASETN')) THEN DELETE;

* Numeric and character dataset attributes; * dataset name, always present;
    %LET Dsattrs = LibName MemLabel Original Modified Engine MemType TypeMem
                   Compress NoDupKey NoDupRec Sorted SortedBy CharSet Collate
                   Physical Logical NofVars;
    SELECT (UPCASE(Attribute));
      WHEN (UPCASE('MemName')) Attribute = &MemName; * dataset name always present;
      %LET I = 1;
      %LET DsAttr = %SCAN ( &DsAttrs, &I, %STR( ) );
      %DO %WHILE (%LENGTH(&DsAttr));
        WHEN (UPCASE("&DsAttr"))
          %IF (%LENGTH(%SUPERQ(&DsAttr))) %THEN
            Attribute = &&&DsAttr;
          %ELSE DELETE;;
        %LET I = %EVAL ( &I + 1 );
        %LET DsAttr = %SCAN ( &DsAttrs, &I, %STR( ) );
      %END;
      OTHERWISE /* unexpected */;
    END;
  RUN;

%if &Debug ge 2 %then %do; %* = = = = = = = = = ;
  data current.MetaDset_Processed; set &MetaDset; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = ;

* = = = = = = = = = ;
* Variable metadata ;
* = = = = = = = = = ;

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
* All variable information is contained in the dataset &Appended.   ;
* Extract variable meta information for each dataset and store it   ;
* in a temporary output dataset MetaVars, one record per attribute  ;
* (VarName, VarType, FmtName, etc.) for all involved datasets.      ;
* Sort the dataset by (variable) Name and dataset number (DatasetN).;
* (All this only if &Variable is not empty.)                        ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Add descriptive statistics per variable to Appended if requested ;
%*__________________________________________________________________;

  %LET NofNames = -1; %* 1st initialisation;
  %IF (%LENGTH(%SUPERQ(Variable))) %THEN %DO;

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
* Limit variables to compare and reference to the optionally    ;
* specified variable list (VarList). This could most simply     ;
* be done by limiting the datasets to extract the CONTENTS      ;
* of to the list using (KEEP=[variable list]). However, not     ;
* all specified variables need to be present in all datasets    ;
* and this might cause the annoying WARN1NG that certain        ;
* variables in the KEEP or DROP list have not been initialised. ;
* In order to avoid those WARN1NGs a different approach is      ;
* taken: the variable list is made into a dataset with variable ;
* names that is (SQL) JOINed with the complete Appended dataset ;
* keeping only those records with the indicated variable names. ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

    %IF (%LENGTH(&VarList)) %THEN %DO; %* &Variable and &VarList;
      DATA &Var_list;
        LENGTH Name $32;
        %LET I = 1;
        %LET VarName = %SCAN(&VarList, &I, %STR( ));
        %DO %WHILE (%LENGTH(&VarName));
          Name = "&VarName";
          OUTPUT;
          %LET I = %EVAL(&I+1);
          %LET VarName = %SCAN(&VarList, &I, %STR( ));
        %END;
      RUN;

      PROC SQL;
        CREATE TABLE &Appended AS
        SELECT A.* FROM &Appended AS A
          INNER JOIN &Var_List AS V
        ON UPCASE(A.Name) EQ UPCASE(V.Name);
      QUIT;

%if &Debug ge 12 %then %do; %* = = = = = = = = = ;
  %PUT === After Var_List and SQL Appended ===;
  data current.Var_List; set &Var_List; run; * debug;
  data current.Appended_SQL; set &Appended; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = ;

    %END; %* ---&VarList specified---;
/*data current.Appended_SQL; set &Appended; run; * debug @@@@@@@@@;*/

%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
%* Recalculate absolute &ComMinNr from percentage or zero  ;
%* or negative value as ratio or difference with &NofDsets.;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
%* Exclude variables that occur less than &ComMinNr times. ;
%* Determine aggregate SQL function N of Name by Name and  ;
%* merge (full join) aggregated result back with original. ;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
%* Similar recalculation and processing for &ComMaxNr      ;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

    %LET NofNames = -1; %* 2nd initialisation;
    %IF (%BQUOTE(&ComMinNr) NE 1 OR %BQUOTE(&ComMaxNr) NE 0) %THEN %DO;
    %* === &Variable and (&ComMinNr or &ComMaxNr) specified ===;
%* No reduction of ComMinNr=1 and ComMaxNr=0 (but recalculation could be done anyway);
      %IF (%QSUBSTR(&ComMinNr,%LENGTH(%SUPERQ(ComMinNr)),1) EQ %STR(%%)) %THEN %DO;
        %LET ComMinNr = %QSUBSTR(&ComMinNr,1,%EVAL(%LENGTH(%SUPERQ(ComMinNr)))-1);
        %LET ComMinNr = %SYSEVALF ( &NofDsets*&ComMinNr/100 );
      %END;
      %ELSE %IF (&ComMinNr LE 0) %THEN
        %LET ComMinNr = %EVAL (&NofDsets + &ComMinNr);
      %IF (%QSUBSTR(&ComMaxNr,%LENGTH(%SUPERQ(ComMaxNr)),1) EQ %STR(%%)) %THEN %DO;
        %LET ComMaxNr = %QSUBSTR(&ComMaxNr,1,%EVAL(%LENGTH(%SUPERQ(ComMaxNr)))-1);
        %LET ComMaxNr = %SYSEVALF ( &NofDsets*&ComMaxNr/100 );
      %END;
      %ELSE %IF (&ComMaxNr LE 0) %THEN
        %LET ComMaxNr = %EVAL (&NofDsets + &ComMaxNr);
      PROC SQL NOPRINT;
        CREATE TABLE &Appended (WHERE=(CommonNr GE &ComMinNr
                                   AND CommonNr LE &ComMaxNr)) AS
          SELECT *
            FROM &Appended AS A          /* original data */
              FULL JOIN
          (SELECT N(Name) AS CommonNr    /* frequency of  */
            FROM &Appended AS B          /* Name per Name */
              GROUP BY Name)
                ON A.Name EQ B.Name;     /* merge by Name */
        SELECT COUNT(*) INTO : NofNames  /* count records left */
          FROM &Appended;
      QUIT; * CommonNr dropped in data step below;

      DATA &Appended;
        SET &Appended (DROP=CommonNr);
      RUN; * (yet dropping CommonNr not really necessary and in the future not done!);
    %END; %* ---ComMinNr and/or ComMaxNr specified, unequal defaults---;

%if &Debug ge 1 %then %do; %* = = = = = = = = = = = = = = = = = = ;
      %put ###DEBUG### after ComMinNr (ComMaxNr) NofNames=&NofNames;
      data current.Appended_ComMinNr; set &Appended; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ;
/*    data current.Appended_ComMinNr; set &Appended; run; * debug @@@@@@@@@;*/

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
* Data step looping through all records (all dataset variables).;
* For each record (unique variable per dataset) do CALL EXECUTE.;
* (Most elegant: call attached auxiliary macro to run delayed.) ;
* Determine descriptive statistics with PROC SQL, while adding. ;
* Finally combine (horizontally) attributes and statistics.     ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

    %IF (&DescStat) %THEN %DO; %* &Variable specified and &DescStat filled;

* Create 2 empty descriptive statistics datasets with Partial structure;
      DATA &StatDesc &SQL_stat;
        LENGTH LibName $8 MemName $32 Name $32;
        STOP;
      RUN;

     %IF (&NofNames) %THEN %DO; /* --- always true ??? --- */
      DATA _NULL_;
        SET &Appended;
        CALL EXECUTE ('%NRSTR(%DescStat (LibName=' || LEFT(TRIM(LibName)) ||
                      ', MemName=' || LEFT(TRIM(MemName)) ||
                      ', Name=' || LEFT(TRIM(Name)) ||
                      ', Type=' || PUT(Type,1.0) ||
                      ', Format=' || LEFT(TRIM(Format)) ||
                      '))' );
      RUN;

      PROC SORT DATA=&Appended; BY LibName MemName Name; RUN;
      PROC SORT DATA=&SQL_stat; BY LibName MemName Name; RUN;
      PROC SORT DATA=&StatDesc; BY LibName MemName Name; RUN;

%if &Debug ge 12 %then %do; %* = = = = = = = = = = = = = = = = = = ;
      %put ###DEBUG### === after CALL EXECUTE ===;
      data current.Appended_CallExec; set &Appended; run; * debug;
      data current.SQL_stat_CallExec; set &SQL_Stat; run; * debug;
      data current.StatDesc_CallExec; set &StatDesc; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ;

      DATA &Appended; * (reorder with RETAIN here of desired) ;
        MERGE &Appended &SQL_stat &StatDesc; BY LibName MemName Name;
      RUN;

%if &Debug ge 12 %then %do; %* = = = = = = = = = = = = = = = = = = ;
      %put ###DEBUG### === after MERGE Appended SQL_Stat StatDesc ===;
      data current.Appended_Merged; set &Appended; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ;

     %END; %* (&NofNames);
    %END; %* (&DescStat);

/* ========= in previous loop, per looping
%if &Debug ge 1 and %sysfunc(exist(&StatDesc)) %then %do; %* = = = = = = = = = = = = = = = = = = ;
     data current.StatDesc; set &StatDesc; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ;
%if &Debug ge 1 %then %do; %* = = = = = = = = = = = = = = = = = = ;
     data current.Appended; set &Appended; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ;
========= */

* Sort &Appended by (variable) Name (if &Variable exists);
   %IF (&NofNames) %THEN %DO; /* --- always true ??? --- */
    PROC SORT DATA=&Appended; BY Name DatasetN; RUN; * safe to add NODUPKEY;

    PROC TRANSPOSE DATA=&Appended (DROP=LibName -- Logical NofVars Type DataFile)                                        
                   OUT =&MetaVars (DROP=_LABEL_)
                   NAME=Attribute PREFIX=Dataset_;
      BY Name;
      ID DatasetN; * together with PREFIX becomes Dataset_#;
      VAR _ALL_; * Numeric and character variables, incl. Name and DatasetN;
* The VAR statement determines the order of Attribute records;
    RUN;

%if &Debug ge 1 %then %do; %* = = = = = = = = = ;
  data current.MetaVars_Transposed; set &MetaVars; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = ;

* Process MetaVars with regard to optional attributes and their names;
    DATA &MetaVars (DROP=Name Remove _I_);
      RETAIN Attribute
      %DO I = 1 %TO &NofDsets;
        Dataset_&I
      %END;
      ; * Define order of variables with RETAIN;
      LENGTH Attribute $32; * (must also contain complete variable names);
      SET &MetaVars;
      AttrPart = 'Variable'; * Constant;
*#####Attribute = UPCASE(Attribute);

* Delete some records that were generated by TRANSPOSing VAR _ALL_;
      IF (UPCASE(Attribute) EQ 'DATASETN') THEN DELETE;
 
* Numeric and character variable attributes; * variable name, always present;
      %IF (%LENGTH(%SUPERQ(VarLabel))) %THEN %DO;
        IF      (UPCASE(Attribute) EQ 'NAME' ) THEN Attribute=&Variable;
        ELSE IF (UPCASE(Attribute) EQ 'LABEL') THEN Attribute=&VarLabel;
      %END;
      %ELSE %DO; * &VarLabel empty (not specified);
        IF      (UPCASE(Attribute) EQ 'NAME' ) THEN DELETE; * Just one name+label record;
        ELSE IF (UPCASE(Attribute) EQ 'LABEL') THEN Attribute = Name;
      %END;
          
      ELSE DO; * other attributes;
        %LET VarAttrs = VarNum Type_Len InFormat Format NofValid Missings
                        Uniques Minimum Fminimum Maximum Fmaximum Mean StDev
                        Median Fmedian Mode Fmode;
        SELECT (UPCASE(Attribute));
          %LET I = 1;
          %LET VarAttr = %SCAN ( &VarAttrs, &I, %STR( ) );
          %DO %WHILE (%LENGTH(&VarAttr));
            WHEN (UPCASE("&VarAttr"))
            %IF (%LENGTH(%SUPERQ(&VarAttr))) %THEN
              Attribute = &&&VarAttr;
            %ELSE DELETE;;
            %LET I = %EVAL ( &I + 1 );
            %LET VarAttr = %SCAN ( &VarAttrs, &I, %STR( ) );
          %END;
          OTHERWISE /* unexpected */;
        END; * --- [SELECT];
* This lets all UNI statistics (in list) pass to be included in CrossRef;
      END; * --- [other atributes];

%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
%* If no numeric statistics then remove attribute/statistic record.  ;
%* This applies to character variables from the datasets referenced. ;
%* Potentially remove record if any Dataset_# is '_' (statistic set).;
%* Do not remove record if any Dataset_# is non-missing and not '_'. ;
%* So only delete record if at least one is '_' and rest is missing, ;
%* not if all are missing or at least one is non-missing and not '_'.;
%* Finally change all '_' values to empty to show as not present.    ;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

      ARRAY Dataset_ [&NofDsets] $; * Existing variables Dataset_# ;
      Remove = -1; * undetermined;
      DO _I_ = 1 TO &NofDsets;
        IF (LEFT(Dataset_[_I_]) EQ '_' AND Remove NE 0) THEN Remove = 1;
        ELSE IF (NOT MISSING(Dataset_[_I_])) THEN Remove = 0;
        IF (LEFT(Dataset_[_I_]) EQ '_') THEN Dataset_[_I_] = '';
      END;
      IF (Remove EQ -1) THEN Remove = 0; * (only all empty values);
      IF (Remove) THEN DELETE;
    RUN;

%if &Debug ge 2 %then %do; %* = = = = = = = = = ;
  data current.MetaVars_Processed; set &MetaVars; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = ;

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
* Combine both (dataset and variable) metadata datasets into one   ;
* Preserve maximum character lengths of all variables by not using ;
* PROC APPEND or a data step, concatenating datasets with SET and  ;
* neither PROC SQL with UNION because that reorders the records in ;
* a very much undesired order, but either macro _Append_  or       ;
* PROC SQL with OUTER UNION CORR (which does not keep the correct  ;
* (in)formats for the columns, but that is no problem here).       ;                                                        ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;

    %*_Append_ (Base=&MetaDset, Data=&MetaVars,
               CalledBy=CrossRef);
    PROC SQL UNDO_POLICY=NONE; /* suppress warning on recursive table references */
      CREATE TABLE &MetaDset AS
        SELECT * FROM &MetaDset
          OUTER UNION CORR
        SELECT * FROM &MetaVars;
    QUIT;

%if &Debug ge 2 %then %do; %* = = = = = = = = = ;
  data current.MetaDset_Combined; set &MetaDset; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = ;

   %END; %* (&NofNames);
  %END; %* ---Variable metadata dependent on existence of &Variable---;

  DATA &CrossRef (DROP=_I_ AttrPart);
    SET &MetaDset;
      BY AttrPart;
* Left justify character converted numerical values;
    ARRAY Dataset_ [&NofDsets] $; * Existing variables Dataset_# ;
    DO _I_ = 1 TO &NofDsets;
      Dataset_[_I_] = LEFT(Dataset_[_I_]);
    END;
    OUTPUT;
    %IF (%LENGTH(%SUPERQ(ColWidth))) %THEN %DO;
      IF (LAST.AttrPart AND AttrPart EQ 'Dataset') THEN DO;
        DO _I_ = 1 TO &NofDsets;
          Dataset_[_I_] = LEFT(PUT(LENGTHC(Dataset_[_I_]),BEST.));
        END;
        Attribute = &ColWidth; * must be between quotes! ;
        OUTPUT;
      END;
    %END;
  RUN;

%*-----;
%Aborting:
%Exit:
%Finish:
%*-----;

%* - - - - - - ;
%* Cleaning up ;
%* - - - - - - ;

%* Restore possibly modified option settings to previous ones;
  %if (&Debug) %then %do;
    OPTIONS &NoMPrint &NoMErr0r &NoSErr0r &NoMLogic &Symbolgn &MacroGen &NoSpool;
  %end;
  OPTIONS VALIDVARNAME = &ValidVar
          LINESIZE=&LineSize PAGESIZE=&PageSize &NoCenter;

%* Delete work datasets;
  %_Delete_ (&Contents &AttrData &Appended &SQL_stat &Uni_stat
             &DsetInfo &MetaDset &MetaVars &StatDesc &Var_List);

%*-------------; %*=======================;
%MEND CrossRef ; %* End of macro CrossRef ;
%*-------------; %*=======================;

%* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *;
%* E n d   o f   C r o s s r e f   (cross reference of multiple datasets) ;
%* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *;

%*==========================; %*----------------;
%* === Auxiliary macros === ; %* Macro DescStat ;
%*==========================; %*----------------;

%MACRO DescStat (LibName=, MemName=, Name=, Type=, Format=);
%* Requested statistics arguments inherited from calling level;
  %LOCAL Fmt /* Format specification dependent on existing format */
         Len /* Length (width) of character fields (i.e. Fmedian) */
  ;
%* Determine format to use;
  %IF (%LENGTH(&Format)) %THEN %LET Fmt = &Format;
  %ELSE %IF (&Type EQ 1) %THEN %LET Fmt = BEST.;
  %ELSE %IF (&Type EQ 2) %THEN %LET Fmt = $CHAR.;
%* Determine total width of format;
  %IF (%LENGTH(&Format)) %THEN %DO;
    %LET Len = %SCAN(&Format,1,.);
    %LET Len = %SYSFUNC(SUBSTRN(&Len,%EVAL(%SYSFUNC(NOTDIGIT(&Len,-%LENGTH(&Len)))+1)));
  %END;
%put LibName=&LibName, MemName=&MemName, Name=&Name;
%put Type=&Type, Format=&Format, Fmt=&Fmt, Len=&Len;
  %IF (NOT %LENGTH(&Len)) %THEN %LET Len = 32; /* fixed, sufficient width */

  PROC SQL UNDO_POLICY=NONE; /* suppress warning on recursive table references */
    CREATE TABLE &SQL_stat (DROP=Dummy) AS /* concatenate SQL_stat */
      SELECT * FROM &SQL_stat              /* with information below */
        OUTER UNION CORR
      SELECT "&LibName" AS LibName,        /* Library name */
             "&MemName" AS MemName,        /* Dataset name */
             "&Name" AS Name               /* Variable name */
        %IF (%LENGTH(&NofValid)) %THEN , N(&Name) AS NofValid;
        %IF (%LENGTH(&Missings)) %THEN , NMISS(&Name) AS Missings;
        %IF (%LENGTH(&Uniques))  %THEN , N(DISTINCT &Name) AS Uniques;
        %IF (%LENGTH(&Minimum))  %THEN %DO; /* always character */
          %IF (&Type EQ 1) %THEN , LEFT(PUT(MIN(&Name),BEST.)) AS Minimum;
          %IF (&Type EQ 2) %THEN , MIN(&Name) AS Minimum;
        %END;
        %IF (%LENGTH(&Fminimum)) %THEN %DO; /* always character */
          , LEFT(PUT(MIN(&Name),&Fmt)) AS Fminimum
        %END;
        %IF (%LENGTH(&Maximum))  %THEN %DO; /* always character */
          %IF (&Type EQ 1) %THEN , LEFT(PUT(MAX(&Name),BEST.)) AS Maximum;
          %IF (&Type EQ 2) %THEN , MAX(&Name) AS Maximum;
        %END;
        %IF (%LENGTH(&Fmaximum)) %THEN %DO; /* always character */
          , LEFT(PUT(MAX(&Name),&Fmt)) AS Fmaximum
        %END;

%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
%* Other statistics should not be included for character variables. ;
%* Statistics never van take any other missing value than ., so     ;
%* indicate statistics to be removed by special missing value ._    ;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;

        %IF (%LENGTH(&Mean)) %THEN %DO;
          %IF (&Type EQ 1) %THEN , MEAN(&Name) AS Mean;
          %ELSE , ._ AS Mean;
        %END;
        %IF (%LENGTH(&StDev)) %THEN %DO;
          %IF (&Type EQ 1) %THEN , STD(&Name) AS StDev;
          %ELSE , ._ AS StDev;
        %END;
        %IF (%LENGTH(&Median)) %THEN %DO;
          , ._ AS Median /* to be overwritten by results of PROC UNIVARIATE */
        %END;
        %IF (%LENGTH(&Fmedian)) %THEN %DO;
          , '_' AS Fmedian LENGTH=&Len /* overwritten by numerical results */
        %END;
        %IF (%LENGTH(&Mode)) %THEN %DO;
          , ._ AS Mode /* to be overwritten by results of PROC UNIVARIATE */
        %END;
        %IF (%LENGTH(&Fmode)) %THEN %DO;
          , '_' AS Fmode LENGTH=&Len /* overwritten by numerical results */
        %END;
          %* and so on .........;

%* SQLstats, user specified list of statistics;
        %LET I = 1;
        %LET SQLstat = %SCAN(&SQLstats, &I, %STR( ));
        %DO %WHILE (%LENGTH(&SQLstat));
          %IF (&Type EQ 1) %THEN , &SQLstat(&Name) AS _&SQLstat;
          %ELSE , ._ AS _&SQLstat; /* predefine every SQLstat as .- to overwrite */
          %LET I = %EVAL ( &I + 1 );
          %LET SQLstat = %SCAN(&SQLstats, &I, %STR( ));
        %END;

%* UNIstats, user specified list of statistics;
        %LET I = 1;
        %LET UNIstat = %SCAN(&UNIstats, &I, %STR( ));
        %DO %WHILE (%LENGTH(&UNIstat));
          , ._ AS _&UNIstat /* predefine every UNIstat as .- to overwrite */
          %LET I = %EVAL ( &I + 1 );
          %LET UNIstat = %SCAN(&UNIstats, &I, %STR( ));
        %END;

%* If no SQL_stat at all then force one aggregated SQL statistic;
        , MAX(.) AS Dummy

    FROM &LibName..&MemName;
  QUIT; * === end of PROC SQL ===;

%if &Debug ge 1 and %sysfunc(exist(&SQL_Stat)) %then %do; %* = = = = = = = = = = = = = = = = = = ;
     data current.SQL_stat; set &SQL_stat; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ;

* Calculation of Median and Mode using PROC UNIVARIATE and merged with SQL_stat;
  %IF ((%LENGTH(&Median)  OR %LENGTH(&Mode) OR
        %LENGTH(&FMedian) OR %LENGTH(&FMode)) AND &Type EQ 1) %THEN %DO;
    PROC UNIVARIATE DATA=&LibName..&MemName NOPRINT;
      VAR &Name;
      OUTPUT OUT=&Uni_stat MEDIAN=Median MODE=Mode
%* UNIstats, user specified list of statistics;
      %LET I = 1;
      %LET UNIstat = %SCAN(&UNIstats, &I, %STR( ));
      %DO %WHILE (%LENGTH(&UNIstat));
        &UNIstat=_&UNIstat
        %LET I = %EVAL ( &I + 1 );
        %LET UNIstat = %SCAN(&UNIstats, &I, %STR( ));
      %END;
      ;
    RUN; * === end of PROC UNIVARIATE ===;
    PROC SQL UNDO_POLICY=NONE; /* suppress warning on recursive table references */
      CREATE TABLE &StatDesc AS        /* concatenate StatDesc */
        SELECT * FROM &StatDesc        /* with information below */
          OUTER UNION CORR
        SELECT "&LibName" AS LibName,  /* Library name */
               "&MemName" AS MemName,  /* Dataset name */
               "&Name" AS Name,        /* Variable name */
               LEFT(PUT(Median,&Fmt)) AS Fmedian, /* formatted */
               LEFT(PUT(Mode,  &Fmt)) AS Fmode  , /* formatted */
               * FROM &Uni_stat;       /* statistics values */
    QUIT; * === end of PROC SQL ===;
  %END;

%if &Debug ge 1 and %sysfunc(exist(&Uni_Stat)) %then %do; %* = = = = = = = = = = = = = = = = = = ;
     data current.Uni_stat; set &Uni_stat; run; * debug;
%end; %* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ;

%*-------------; %*=======================;
%MEND DescStat ; %* End of macro DescStat ;
%*-------------; %*=======================;
